/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.matching;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Writer;
import java.util.StringTokenizer;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import profi.model.Point;
import profi.model.Polyline;

public class Relation {

	public static final int PLACE_CENTER	= 1001;
	public static final int PLACE_INSIDE	= 1002;
	public static final int PLACE_INSIDE_B	= 1003;
	public static final int PLACE_BOUNDARY	= 1004;
	public static final int PLACE_OUTSIDE_B	= 1005;
	public static final int PLACE_NEIGHBOUR	= 1006;
	public static final int PLACE_DEFAULT	= 1007;
	
	
	public static final int REL_SAME		= 2000;
	public static final int REL_ROTATED		= 2001;
	public static final int REL_MIRROR		= 2002;
	public static final int REL_DEFAULT		= 2999;
		

	public static Relation readData_01(BufferedReader reader) throws IOException {
		Relation rel = new Relation();
		
		String line;
		String[] parts;
		line = reader.readLine().trim();
		parts = line.split(" = ");
		rel.f1 = Integer.parseInt(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		rel.f2 = Integer.parseInt(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		rel.weightAbs = Double.parseDouble(parts[1]);
		//line = reader.readLine().trim();
		//parts = line.split(" = ");
		//rel.place = Integer.parseInt(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		rel.rel = Integer.parseInt(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		rel.centerDistAbs = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		rel.centerDistRel = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		rel.ratio = Double.parseDouble(parts[1]);
		reader.readLine();
		return rel;
	}

	public static Relation readData_02(BufferedReader reader) throws IOException {
		Relation rel = new Relation();
		
		String line;
		String field;
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		rel.f1 = Integer.parseInt(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		rel.f2 = Integer.parseInt(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		rel.weightAbs = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		rel.rel = Integer.parseInt(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		rel.centerDistAbs = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		rel.centerDistRel = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		rel.ratio = Double.parseDouble(field);
		reader.readLine();
		return rel;
	}

	public static Relation readData_03(Element element) {
		Relation rel = new Relation();
		String type = element.getAttributes().getNamedItem("type").getNodeValue();
		if (type.equalsIgnoreCase("same")) {
			rel.rel = Relation.REL_SAME;
		} else if (type.equalsIgnoreCase("rotated")) {
			rel.rel = Relation.REL_ROTATED;
		} else if (type.equalsIgnoreCase("mirror")) {
			rel.rel = Relation.REL_MIRROR;
		} else if (type.equalsIgnoreCase("none")) {
			rel.rel = Relation.REL_DEFAULT;
		} else {
			throw new RuntimeException();
		}
		rel.f2 = Integer.parseInt(element.getAttributes().getNamedItem("pos").getNodeValue());
		
		NodeList childNodes = element.getChildNodes();
		int nodeCount = 0;
		int stepCount = 0;
		while (stepCount <= 3) {
			Node node = childNodes.item(nodeCount);
			if (node.getNodeType() == Node.ELEMENT_NODE) {
				Element e = (Element) node;
				switch (stepCount) {
				case 0:
					rel.weightAbs = Double.parseDouble(e.getTextContent());
					break;
				case 1:
					rel.centerDistAbs = Double.parseDouble(e.getTextContent());
					break;
				case 2:
					rel.centerDistRel = Double.parseDouble(e.getTextContent());
					break;
				case 3:
					rel.ratio = Double.parseDouble(e.getTextContent());
					break;
				}
				stepCount++;
			}
			nodeCount++;
		}
		return rel;
	}
	
	public double weightAbs;
	public double weightRel;
	
	//public int place;
	public int rel;
	
	public double centerDistAbs;
	public double centerDistRel;
	//public double angle;
	public double ratio;
	
	public int f1;
	public int f2;
	
	public Relation partner;
	
	private Relation() {		
	}
	
	public Relation(Figure first, Figure second) {
		weightAbs = 1.0;
		this.f1 = first.idx;
		this.f2 = second.idx;
	}
	
	public Relation(int first, int second) {
		weightAbs = 1.0;
		this.f1 = first;
		this.f2 = second;
	}

	public void writeData_01(Writer writer, String prefix) throws IOException {
		String pprefix = prefix + "\t";
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("RELATION begin\n");
		buff.append(pprefix); buff.append("f1 = "); buff.append(f1); buff.append("\n");
		buff.append(pprefix); buff.append("f2 = "); buff.append(f2); buff.append("\n");		
		buff.append(pprefix); buff.append("weight = "); buff.append(weightAbs); buff.append("\n");
		//buff.append(pprefix); buff.append("place = "); buff.append(place); buff.append("\n");
		buff.append(pprefix); buff.append("rel = "); buff.append(rel); buff.append("\n");
		buff.append(pprefix); buff.append("distAbs = "); buff.append(centerDistAbs); buff.append("\n");
		buff.append(pprefix); buff.append("distRel = "); buff.append(centerDistRel); buff.append("\n");
		buff.append(pprefix); buff.append("ratio = "); buff.append(ratio); buff.append("\n");
		
		buff.append(prefix); buff.append("RELATION end\n");
		writer.write(buff.toString());
	}

	public void writeData_02(Writer writer, String prefix) throws IOException {
		String pprefix = prefix + "\t";
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("<RELATION>\n");
		buff.append(pprefix); buff.append("<f1>"); buff.append(f1); buff.append("</f1>\n");
		buff.append(pprefix); buff.append("<f2>"); buff.append(f2); buff.append("</f2>\n");		
		buff.append(pprefix); buff.append("<weight>"); buff.append(weightAbs); buff.append("</weight>\n");
		buff.append(pprefix); buff.append("<rel>"); buff.append(rel); buff.append("</rel>\n");
		buff.append(pprefix); buff.append("<distAbs>"); buff.append(centerDistAbs); buff.append("</distAbs>\n");
		buff.append(pprefix); buff.append("<distRel>"); buff.append(centerDistRel); buff.append("</distRel>\n");
		buff.append(pprefix); buff.append("<ratio>"); buff.append(ratio); buff.append("</ratio>\n");
		
		buff.append(prefix); buff.append("</RELATION>\n");
		writer.write(buff.toString());
	}

	public void writeData_03(Writer writer, String prefix) throws IOException {
		String pprefix = prefix + "\t";
		StringBuffer buff = new StringBuffer();
		buff.append(prefix);
		buff.append("<relation");
		buff.append(" type=\"");
		buff.append(getTypeName());
		buff.append("\"");
		buff.append(" pos=\"");
		buff.append(f2);
		buff.append("\"");				
		buff.append(">\n");		
		buff.append(pprefix); buff.append("<weight>"); buff.append(weightAbs); buff.append("</weight>\n");
		buff.append(pprefix); buff.append("<distAbs>"); buff.append(centerDistAbs); buff.append("</distAbs>\n");
		buff.append(pprefix); buff.append("<distRel>"); buff.append(centerDistRel); buff.append("</distRel>\n");
		buff.append(pprefix); buff.append("<ratio>"); buff.append(ratio); buff.append("</ratio>\n");
		buff.append(prefix); buff.append("</relation>\n");
		writer.write(buff.toString());
	}
	
	public String getTypeName() {
		switch (rel) {
			case REL_SAME:
				return "same";
			case REL_ROTATED:
				return "rotated";
			case REL_MIRROR:
				return "mirror";
			case REL_DEFAULT:
				return "none";
			default:
				throw new RuntimeException();
		}
	}
	
	public double vote(Relation other) {
		
		double maxRatio = Math.max(this.ratio, other.ratio);
		double minRatio = Math.min(this.ratio, other.ratio);
		double maxDist  = Math.max(this.centerDistAbs, other.centerDistAbs);
		double minDist  = Math.min(this.centerDistAbs, other.centerDistAbs);
				
		return 
				votePlace(other) * 
				voteRel(other) *
				Math.sqrt(minRatio / maxRatio);		
	}

	public double votePlace(Relation other) {
		/*
		switch (this.place) {
			case PLACE_CENTER: 		switch (other.place) {
										case PLACE_CENTER:		return 1.0;
										case PLACE_INSIDE:		return 0.9;
										case PLACE_INSIDE_B:	return 0.75;
										case PLACE_BOUNDARY:	return 0.7;
										case PLACE_OUTSIDE_B:	return 0.5;
										case PLACE_NEIGHBOUR:	return 0.25;
										case PLACE_DEFAULT:		return 0.5;
										default: throw new RuntimeException();
									}
			case PLACE_INSIDE:		switch (other.place) {
										case PLACE_CENTER:		return 0.9;
										case PLACE_INSIDE:		return 1.0;
										case PLACE_INSIDE_B:	return 0.9;
										case PLACE_BOUNDARY:	return 0.75;
										case PLACE_OUTSIDE_B:	return 0.6;
										case PLACE_NEIGHBOUR:	return 0.3;
										case PLACE_DEFAULT:		return 0.5;
										default: throw new RuntimeException();
									}
			case PLACE_INSIDE_B:	switch (other.place) {
										case PLACE_CENTER:		return 0.75;
										case PLACE_INSIDE:		return 0.9;
										case PLACE_INSIDE_B:	return 1.0;
										case PLACE_BOUNDARY:	return 0.9;
										case PLACE_OUTSIDE_B:	return 0.8;
										case PLACE_NEIGHBOUR:	return 0.5;
										case PLACE_DEFAULT:		return 0.5;
										default: throw new RuntimeException();
									}
			case PLACE_BOUNDARY:	switch (other.place) {
										case PLACE_CENTER:		return 0.7;
										case PLACE_INSIDE:		return 0.75;
										case PLACE_INSIDE_B:	return 0.9;
										case PLACE_BOUNDARY:	return 1.0;
										case PLACE_OUTSIDE_B:	return 0.9;
										case PLACE_NEIGHBOUR:	return 0.6;
										case PLACE_DEFAULT:		return 0.5;
										default: throw new RuntimeException();
									}
			case PLACE_OUTSIDE_B:	switch (other.place) {
										case PLACE_CENTER:		return 0.5;
										case PLACE_INSIDE:		return 0.6;
										case PLACE_INSIDE_B:	return 0.8;
										case PLACE_BOUNDARY:	return 0.9;
										case PLACE_OUTSIDE_B:	return 1.0;
										case PLACE_NEIGHBOUR:	return 0.75;
										case PLACE_DEFAULT:		return 0.5;
										default: throw new RuntimeException();
									}
			case PLACE_NEIGHBOUR:	switch (other.place) {
										case PLACE_CENTER:		return 0.25;
										case PLACE_INSIDE:		return 0.3;
										case PLACE_INSIDE_B:	return 0.5;
										case PLACE_BOUNDARY:	return 0.6;
										case PLACE_OUTSIDE_B:	return 0.75;
										case PLACE_NEIGHBOUR:	return 1.0;
										case PLACE_DEFAULT:		return 0.5;
										default: throw new RuntimeException();
									}
			case PLACE_DEFAULT:	switch (other.place) {
										case PLACE_CENTER:		return 0.5;
										case PLACE_INSIDE:		return 0.5;
										case PLACE_INSIDE_B:	return 0.5;
										case PLACE_BOUNDARY:	return 0.5;
										case PLACE_OUTSIDE_B:	return 0.5;
										case PLACE_NEIGHBOUR:	return 0.5;
										case PLACE_DEFAULT:		return 1.0;
										default: throw new RuntimeException();
		}
			
			default: throw new RuntimeException();
		}
		*/
		double dd = Math.abs(this.centerDistRel - other.centerDistRel);
		double ddd = dd / Math.max(1.0, Math.min(this.centerDistRel, other.centerDistRel));
		return 0.75*Math.pow(1.5, -ddd)+0.25;
	}
	
	
	public double voteRel(Relation other) {

		switch (this.rel) {
			case REL_SAME: 		switch (other.rel) {
										case REL_SAME:		return 1.0;
										case REL_ROTATED:	return 0.9;
										case REL_MIRROR:	return 0.9;
										case REL_DEFAULT:	return 0.5;
										default: throw new RuntimeException();
								}
			case REL_ROTATED: 	switch (other.rel) {
										case REL_SAME:		return 0.9;
										case REL_ROTATED:	return 1.0;
										case REL_MIRROR:	return 0.9;
										case REL_DEFAULT:	return 0.5;
										default: throw new RuntimeException();
										
								}
			case REL_MIRROR: 	switch (other.rel) {
										case REL_SAME:		return 0.9;
										case REL_ROTATED:	return 0.9;
										case REL_MIRROR:	return 1.0;
										case REL_DEFAULT:	return 0.5;
										default: throw new RuntimeException();
								}
			case REL_DEFAULT: 	switch (other.rel) {
										case REL_SAME:		return 0.5;
										case REL_ROTATED:	return 0.5;
										case REL_MIRROR:	return 0.5;
										case REL_DEFAULT:	return 1.0;
										default: throw new RuntimeException();
								}
			default: throw new RuntimeException();
		}
	}	
}
