/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.matching;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Writer;
import java.util.StringTokenizer;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import profi.model.Point;
import profi.model.Polyline;
import profi.model.ShapeSet;
import profi.model.SimilarityTransformation;
import profi.model.Transformation;

public class Triangle extends Figure {

	static Triangle createStandardFigure() {
		Triangle tri = new Triangle();
		tri.p1 = new Point(0.0, 0.0);
		tri.p2 = new Point(100.0, 0.0);
		tri.p3 = new Point(50.0, 86.6);
		tri.length = tri.p1.distance(tri.p2) + tri.p2.distance(tri.p3) + tri.p3.distance(tri.p1);
		return tri;
	}
	
	private Point p1;
	private Point p2;
	private Point p3;
	private double length;
	
	public Triangle() {
		super(Figure.TRIANGLE);
	}

	public Triangle(Point p1, Point p2, Point p3) {
		super(Figure.TRIANGLE);
		this.p1 = new Point(p1);
		this.p2 = new Point(p2);
		this.p3 = new Point(p3);
		length = p1.distance(p2) + p2.distance(p3) + p3.distance(p1);		
	}
	
	public Triangle(Triangle other) {
		super(other);
		p1 = new Point(other.p1);
		p2 = new Point(other.p2);
		p3 = new Point(other.p3);
		length = other.length;
	}

	public void addSpecificData_01(Writer writer, String prefix) throws IOException {
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("TRIANGLE\n");
		buff.append(prefix); buff.append("p1.x = "); buff.append(p1.x); buff.append("\n");
		buff.append(prefix); buff.append("p1.y = "); buff.append(p1.y); buff.append("\n");
		buff.append(prefix); buff.append("p2.x = "); buff.append(p2.x); buff.append("\n");
		buff.append(prefix); buff.append("p2.y = "); buff.append(p2.y); buff.append("\n");
		buff.append(prefix); buff.append("p3.x = "); buff.append(p3.x); buff.append("\n");
		buff.append(prefix); buff.append("p3.y = "); buff.append(p3.y); buff.append("\n");
		writer.write(buff.toString());
	}
	
	public void addSpecificData_02(Writer writer, String prefix) throws IOException {
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("<!-- TRIANGLE -->\n");
		buff.append(prefix); buff.append("<p1.x>"); buff.append(p1.x); buff.append("</p1.x>\n");
		buff.append(prefix); buff.append("<p1.y>"); buff.append(p1.y); buff.append("</p1.y>\n");
		buff.append(prefix); buff.append("<p2.x>"); buff.append(p2.x); buff.append("</p2.x>\n");
		buff.append(prefix); buff.append("<p2.y>"); buff.append(p2.y); buff.append("</p2.y>\n");
		buff.append(prefix); buff.append("<p3.x>"); buff.append(p3.x); buff.append("</p3.x>\n");
		buff.append(prefix); buff.append("<p3.y>"); buff.append(p3.y); buff.append("</p3.y>\n");
		writer.write(buff.toString());
	}
	
	public void readSpecificData_01(BufferedReader reader) throws IOException {
		String line;
		String[] parts;
		p1 = new Point();
		p2 = new Point();
		p3 = new Point();
		
		line = reader.readLine().trim();
		parts = line.split(" = ");
		p1.x = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		p1.y = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		p2.x = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		p2.y = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		p3.x = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		p3.y = Double.parseDouble(parts[1]);
		length = p1.distance(p2) + p2.distance(p3) + p3.distance(p1);		
		do {
			line = reader.readLine().trim();
		} while (!line.equalsIgnoreCase("FIGURE end"));
	}

	public void readSpecificData_02(BufferedReader reader) throws IOException {
		String line;
		String field;
		p1 = new Point();
		p2 = new Point();
		p3 = new Point();
		
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		p1.x = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		p1.y = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		p2.x = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		p2.y = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		p3.x = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		p3.y = Double.parseDouble(field);
		length = p1.distance(p2) + p2.distance(p3) + p3.distance(p1);		
		do {
			line = reader.readLine().trim();
		} while (!line.equalsIgnoreCase("</FIGURE>"));
	}
	
	public ShapeSet getShapeSet() {
		ShapeSet set = new ShapeSet();
		double a = p2.distance(p3);
		double b = p3.distance(p1);
		double c = p1.distance(p2);
		double cx = (a*p1.x + b*p2.x + c*p3.x) / (a + b + c);
		double cy = (a*p1.y + b*p2.y + c*p3.y) / (a + b + c);
		for (int l = -1; l < layers.size(); l++) {
			double r = 1.0;
			if (l > -1) {
				r = ((Double) layers.get(l)).doubleValue();
			}
			Polyline poly = new Polyline(true);
			double x, y;
			x = r*p1.x + (1-r)*cx;
			y = r*p1.y + (1-r)*cy;
			poly.addPoint(new Point(x, y));
			x = r*p2.x + (1-r)*cx;
			y = r*p2.y + (1-r)*cy;
			poly.addPoint(new Point(x, y));
			x = r*p3.x + (1-r)*cx;
			y = r*p3.y + (1-r)*cy;
			poly.addPoint(new Point(x, y));
			set.add(poly);
		}
		return set;
	}

	public double getSize() {
		return length / 2;
	}

	public double getSalience() {
		return getSize() * (0.25 + maybeFrame());
	}

	public double getLength() {
		return length;
	}

	public Point getCenter() {
		return new Point((p1.x + p2.x + p3.x) / 3, (p1.y + p2.y + p3.y) / 3);
	}

	public Figure copy() {
		return new Triangle(this);
	}

	public boolean isCopy(Figure other) {
		if (other instanceof Triangle) return isTheSame((Triangle) other);
		else return false;
	}	

	public double sqDist(Figure o) {
		if (!(o instanceof Triangle)) {
			return Double.MAX_VALUE;
		}
		Triangle other = (Triangle) o;
		Point p11 = this.p1;
		Point p12 = this.p2;
		Point p13 = this.p3;
		Point p21 = other.p1;
		Point p22 = other.p2;
		Point p23 = other.p3;
		double d11 = p11.distanceSQ(p21);
		double d12 = (p11.x - p22.x) * (p11.x - p22.x) + (p11.y - p22.y) * (p11.y - p22.y);
		double d13 = (p11.x - p23.x) * (p11.x - p23.x) + (p11.y - p23.y) * (p11.y - p23.y);
		double d21 = (p12.x - p21.x) * (p12.x - p21.x) + (p12.y - p21.y) * (p12.y - p21.y);
		double d22 = (p12.x - p22.x) * (p12.x - p22.x) + (p12.y - p22.y) * (p12.y - p22.y);
		double d23 = (p12.x - p23.x) * (p12.x - p23.x) + (p12.y - p23.y) * (p12.y - p23.y);
		double d31 = (p13.x - p21.x) * (p13.x - p21.x) + (p13.y - p21.y) * (p13.y - p21.y);
		double d32 = (p13.x - p22.x) * (p13.x - p22.x) + (p13.y - p22.y) * (p13.y - p22.y);
		double d33 = (p13.x - p23.x) * (p13.x - p23.x) + (p13.y - p23.y) * (p13.y - p23.y);
		double d1 = Math.min(d11, Math.min(d12, d13));
		double d2 = Math.min(d21, Math.min(d22, d23));
		double d3 = Math.min(d31, Math.min(d32, d33));
		return Math.max(d1, Math.max(d2, d3));
	}
	
	public boolean isTheSame(Triangle other) {
		double maxRadius = Math.max(this.length, other.length) / 4;
		double maxSqDist = sqDist(other);
		return (maxSqDist < maxRadius * maxRadius / 5 / 5);
	}


	public void transfor(Transformation t) {
		if (t instanceof SimilarityTransformation) {
			apply((SimilarityTransformation) t);
		} else {
			t.applyTo(this.p1);
			t.applyTo(this.p2);
			t.applyTo(this.p3);
			this.length = Math.sqrt(Math.max(p1.distanceSQ(p2), Math.max(p1.distanceSQ(p3), p2.distanceSQ(p3))));
		}		
	}
	
	public void apply(SimilarityTransformation sim) {
		sim.applyTo(this.p1);
		sim.applyTo(this.p2);
		sim.applyTo(this.p3);
		this.length *= sim.getScale();
	}
			
	public double getLengthSq(int i) {
		Point p, q;
		switch (i) {
			case 0: p = p1; q = p2; break;
			case 1: p = p2; q = p3; break;
			case 2: p = p3; q = p1; break;
			default: p = null; q = null;
		}		
		return p.distanceSQ(q);
	}

	public double weightFactor() {
		return 1.5;
	}

	public boolean layering(Figure fig) {
		if (fig instanceof Triangle) {
			Triangle other = (Triangle) fig;
			
			double a1 = p2.distance(p3);
			double b1 = p3.distance(p1);
			double c1 = p1.distance(p2);
			double cx1 = (a1*p1.x + b1*p2.x + c1*p3.x) / (a1 + b1 + c1);
			double cy1 = (a1*p1.y + b1*p2.y + c1*p3.y) / (a1 + b1 + c1);
			double s1 = (a1+b1+c1)/2.0;
			double area1 = Math.sqrt(s1*(s1-a1)*(s1-b1)*(s1-c1));

			double a2 = other.p2.distance(other.p3);
			double b2 = other.p3.distance(other.p1);
			double c2 = other.p1.distance(other.p2);
			double m2 = Math.max(a2, Math.max(b2, c2));
			
			Point[] p1 = {this.p1, this.p2, this.p3};
			Point[] p2 = {other.p1, other.p2, other.p3};
			int[] perm = {0, 1, 2};
			double dist = Double.MAX_VALUE;
			
			for (int i = 0; i < 3; i++) {
				for (int j = 0; j < 3; j++) {
					if (j==i) continue;
					int k = 3 - j - i;
					double d0 = p1[0].distanceSQ(p2[i]);
					double d1 = p1[1].distanceSQ(p2[j]);
					double d2 = p1[2].distanceSQ(p2[k]);
					double d = d0 + d1 + d2;
					if (d < dist) {
						dist = d;
						perm[0] = i;
						perm[1] = j;
						perm[2] = k;
					}
				}
			}
			
			// evaluating r by minimizing least squares;
			double nom = 0.0;
			double den = 0.0;
			for (int i = 1; i < 3; i++) {
				nom += (cx1-p2[perm[i]].x)*(cx1-p1[i].x) + (cy1-p2[perm[i]].y)*(cy1-p1[i].y);
				den += (cx1-p1[i].x)*(cx1-p1[i].x) + (cy1-p1[i].y)*(cy1-p1[i].y);
			}
			
			if (Math.abs(den) < 1e-6) return false;
			
			double rNew = nom / den;
			double diff = 0.0;
			for (int i = 1; i < 3; i++) {
				diff += p2[perm[i]].distanceSQ(new Point(
						rNew*p1[i].x + (1-rNew)*cx1,
						rNew*p1[i].y + (1-rNew)*cy1));
			}
			diff = Math.sqrt(diff/3);
			if (rNew < 0.5 || rNew > 1.0 || diff > Math.sqrt(area1*2) * rNew / 5) {
				return false;
			}
			
			for (int l = -1; l < layers.size(); l++) {
				double r = 1.0;
				if (l > -1) {
					r = ((Double) layers.get(l)).doubleValue();
				}
				if(Math.abs(rNew-r) < 0.2) {
					return true;
				}
			}
			layers.add(new Double(rNew));
			return true;
		}
		return false;
	}
	
	public double maybeFrame() {
		double l1 = p1.distance(p2);
		double l2 = p2.distance(p3);
		double l3 = p3.distance(p1);
		double lMin = Math.min(l1, Math.min(l2, l3));
		double lMax = Math.max(l1, Math.max(l2, l3));
		double ldmax = (lMax-lMin) / lMax;
		double ldmin = Math.min(Math.abs(l1-l2), Math.min(Math.abs(l2-l3), Math.abs(l3-l1))) / lMax;
		return Math.max(0.0, 1.0 - ldmax*ldmin*3);
	}

	public double pithiness() {
		return maybeFrame();
	}
	
	public void addSpecificData_03(Writer writer, String prefix) throws IOException {
		String pprefix = prefix + "\t";
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("<triangle>\n");
		buff.append(pprefix); buff.append("<p1>"); buff.append(p1.x); buff.append(" "); buff.append(p1.y); buff.append("</p1>\n");
		buff.append(pprefix); buff.append("<p2>"); buff.append(p2.x); buff.append(" "); buff.append(p2.y); buff.append("</p2>\n");
		buff.append(pprefix); buff.append("<p3>"); buff.append(p3.x); buff.append(" "); buff.append(p3.y); buff.append("</p3>\n");
		buff.append(prefix); buff.append("</triangle>\n");
		writer.write(buff.toString());		
	}

	public String getTypeName() {
		return "triangle";
	}

	public void readSpecificData_03(Element element) {
		NodeList childNodes = element.getChildNodes();
		int nodeCount = 0;
		int stepCount = 0;
		while (stepCount <= 2) {
			Node node = childNodes.item(nodeCount);
			if (node.getNodeType() == Node.ELEMENT_NODE) {
				Element e = (Element) node;
				switch (stepCount) {
				case 0:
					StringTokenizer tokens = new StringTokenizer(e.getTextContent());
					p1 = new Point(Double.parseDouble(tokens.nextToken()), Double.parseDouble(tokens.nextToken()));
					break;
				case 1:
					tokens = new StringTokenizer(e.getTextContent());
					p2 = new Point(Double.parseDouble(tokens.nextToken()), Double.parseDouble(tokens.nextToken()));
					break;
				case 2:
					tokens = new StringTokenizer(e.getTextContent());
					p3 = new Point(Double.parseDouble(tokens.nextToken()), Double.parseDouble(tokens.nextToken()));
					break;
				}
				stepCount++;
			}
			nodeCount++;
		}
	}

	public Object getItemSource() {
		// TODO Auto-generated method stub
		return null;
	}

	public void setItemSource(Object source) {
		// TODO Auto-generated method stub
		
	}
}
