/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.model;

public class Box {

	private double x1, y1, x2, y2;

	/**
	 * @param x1
	 * @param y1
	 * @param x2
	 * @param y2
	 */
	public Box(double x1, double y1, double x2, double y2) {
		this.x1 = Math.min(x1, x2);
		this.y1 = Math.min(y1, y2);
		this.x2 = Math.max(x1, x2);
		this.y2 = Math.max(y1, y2);
	}

	public Point getCenter() {
		return new Point((x1 + x2) / 2, (y1 + y2) / 2);
	}

	public double getWidth() {
		return x2 - x1;
	}

	public double getHeight() {
		return y2 - y1;
	}

	public double getMaxDim() {
		return Math.max(getWidth(), getHeight());
	}
	
	public double getDiag() {
		return Math.sqrt(getWidth()*getWidth() + getHeight()*getHeight());
	}
	
	public double getDiam() {
		return 2 * (getWidth() + getHeight());
	}

	public double getMinX(){
		return x1;
	}
	
	public double getMaxX(){
		return x2;
	}
	
	public double getMinY(){
		return y1;
	}
	
	public double getMaxY(){
		return y2;
	}
	
	public void add(Box b) {
		x1 = Math.min(x1, b.x1);
		y1 = Math.min(y1, b.y1);
		x2 = Math.max(x2, b.x2);
		y2 = Math.max(y2, b.y2);
	}
}
