/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.segmentation;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;

import javax.imageio.ImageIO;

public class PixelImage {

	public static int background = 255;
	
	public static void test() {
		System.out.println("...");		
	}

	public static int getOther() {
		return 255 - background;
	}

	public static int getOther(int color) {
		return 255 - color;
	}
	
	public static PixelImage getPixelImage(String str) {
		try {
			File file = new File(str);
			BufferedImage img = ImageIO.read(file);
			PixelImage pi = new PixelImage(img);
			pi.name = file.getName();
			return pi;
		} catch (IOException ex) {
			System.err.println(str);
			ex.printStackTrace();
		}
		return null;
	}
	
	public String name;
	private BufferedImage img;
	private int height;
	private int width;
	public int bg; 
	
	public PixelImage(BufferedImage img) {
		name = "'no name'";
		this.img = img;
		this.height = img.getHeight() + 2;
		this.width = img.getWidth() + 2;
		this.bg = background;
		int corners = 0;
		corners += (this.getBrightness(1, 1) == getOther()) ? 1 : 0;
		corners += (this.getBrightness(width-2, 1) == getOther()) ? 1 : 0;
		corners += (this.getBrightness(width-2, height-2) == getOther()) ? 1 : 0;
		corners += (this.getBrightness(1, height-2) == getOther()) ? 1 : 0;
		
		corners +=  (this.getBrightness(width/2, 1) == getOther()) ? 1 : 0;
		corners += (this.getBrightness(width/2, height-2) == getOther()) ? 1 : 0;
		corners += (this.getBrightness(width-2, height/2) == getOther()) ? 1 : 0;
		corners += (this.getBrightness(1, height/2) == getOther()) ? 1 : 0;
		
		if (corners >= 5) {
			this.bg = getOther();
		}
	}
	
	public BufferedImage getImage() {
		return img;
	}
	
	public int getWidth() {
		return this.width;
	}
	
	public int getHeight() {
		return this.height;
	}
	
	public int getOriginalWidth() {
		return this.img.getWidth();
	}
	
	public int getOriginalHeight() {
		return this.img.getHeight();
	}
	
	public int getBrightness(int x, int y) {
		if (x < 1 || x >= this.width - 1 || y < 1 || y >= this.height - 1) {
			return bg;
		}		
		int pixel = this.img.getRGB(x-1, y-1);
        int red   = (pixel >> 16) & 0xff;
        int green = (pixel >>  8) & 0xff;
        int blue  = (pixel      ) & 0xff;
        return ((red + green + blue) / 3 > 170) ? 255 : 0;
	}
	
	public boolean isOther(int x, int y, int ref) {
		return getBrightness(x, y) / 128 != ref / 128; 
	}
}
